/*
 *  Copyright (c) 1996-2007 Digi International Inc., All Rights Reserved
 *
 *  This software contains proprietary and confidential information of Digi
 *  International Inc.  By accepting transfer of this copy, Recipient agrees
 *  to retain this software in confidence, to prevent disclosure to others,
 *  and to make no use of this software other than that for which it was
 *  delivered.  This is an unpublished copyrighted work of Digi International
 *  Inc.  Except as permitted by federal law, 17 USC 117, copying is strictly
 *  prohibited.
 *
 *  Restricted Rights Legend
 *
 *  Use, duplication, or disclosure by the Government is subject to
 *  restrictions set forth in sub-paragraph (c)(1)(ii) of The Rights in
 *  Technical Data and Computer Software clause at DFARS 252.227-7031 or
 *  subparagraphs (c)(1) and (2) of the Commercial Computer Software -
 *  Restricted Rights at 48 CFR 52.227-19, as applicable.
 *
 *  Digi International Inc. 11001 Bren Road East, Minnetonka, MN 55343
 *
 * Description.
 * =======================================================================
 *  tx_application_define() is called just before the ThreadX kernel loads.
 *  It is responsible for creating at least one thread to drive the application.
 *  We create a thread that starts up NET+OS.
 *
 *
 *
 * Edit Date/Ver   Edit Description
 * ==============  =======================================================
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <tx_api.h>
#include <hservapi.h>
#include <sysAccess.h>
#include <bsp_api.h>
#include <sockapi.h>
#include <ssh.h>
#include "init.h"
#include <cliapi.h>
#include <appconf.h>

/* Cleans up spawned thread resources */
#define CLEANUP_STACK_SIZE 0x2000
#define CLEANUP_QUEUE_LENGTH 64
#define CMD_CLEANUP 0x01
static char cleanup_queue_messages[CLEANUP_QUEUE_LENGTH];
static TX_QUEUE cleanup_queue;


#define FORCE_USE_OF_STATIC_KEYS        (1)
#define USE_STATIC_KEYS_IF_FS_DISABLED  (NABspFilesystemForCLib == 0)

/*
 * Set SHOULD_USE_STATIC_KEYS to FORCE_USE_OF_STATIC_KEYS to force this program
 * to use the keys staticly defined in this file.  Set SHOULD_USE_STATIC_KEYS
 * to USE_STATIC_KEYS_IF_FS_DISABLED to use keys in the file system if the
 * file system has been enabled.
 */
#define SHOULD_USE_STATIC_KEYS          USE_STATIC_KEYS_IF_FS_DISABLED


/*
 *
 *  Function: static void ssh_app_thread_cleanup (unsigned long)
 *
 *  Description:
 * 
 *      This function defines a thread that reads 
 *      stack and thread pointers sent from
 *      threads requesting clean up, via a message 
 *      queue.  Thread are removed from the thread list,
 *      and stack/thread pointers are freed if provided.
 *
 *  Parameters:
 *
 *      none
 *
 *  Return Values:
 *
 *      none
 * 
 */ 
static void ssh_app_thread_cleanup(unsigned long param)
{
    unsigned long message[4];
    while(1)
    {
        tx_thread_sleep(10);

        /* wait for a command */
        if(tx_queue_receive(&cleanup_queue, message, TX_WAIT_FOREVER) != TX_SUCCESS)
        {
            tx_thread_sleep(10);
            continue;
        }

        switch(message[0])
        {
            case CMD_CLEANUP:
                /* terminate thread */
                if(message[1] != 0)
                {
                    tx_thread_terminate((TX_THREAD*)message[1]);
                    tx_thread_delete((TX_THREAD*)message[1]);
                    if(message[2] != 0)
                        free((void*)message[2]);
                    free((void*)message[1]);
                }
                break;
            default:
                break;
        }
    }     
}

/*
 *
 *  Function: static void ssh_app_send_cleanup (void)
 *
 *  Description:
 * 
 *     This function is used to clean up thread
 *     resources from spawned processes. This sends
 *     a clean up message to the clean up queue, which 
 *     is received and processed by the clean up thread.
 *     Threads that call this function will be suspended,
 *     and their thread / stack pointers will be freed  
 *
 *  Parameters:
 *
 *      none
 *
 *  Return Values:
 *
 *      none
 * 
 */ 
void ssh_app_send_cleanup()
{
    unsigned long message[4];
    TX_THREAD* cur_thread;
  
    cur_thread = tx_thread_identify();
    message[0] = CMD_CLEANUP;
    message[1] = (unsigned long)cur_thread;
    message[2] = (unsigned long)cur_thread->tx_stack_start;
    tx_queue_send(&cleanup_queue, message, TX_WAIT_FOREVER);
    tx_thread_suspend(cur_thread);
    return;
}

/*
 *
 *  Function: static int readKeysFromDB (char*, char*, int)
 *
 *  Description:
 * 
 *     This function reads up to buflen bytes of the specified file 
 *     from the filesystem into the specified buffer. 
 *
 *  Parameters:
 *
 *      char* fpath : filesystem path to the file
 *      char* buf   : a buffer to read the file contents into
 *      int buflen  : the length of buf 
 *
 *  Return Values:
 *
 *      int  the number of bytes read
 * 
 */ 
static int readKeysFromDB(char *fpath, char *buf, int buflen)
{
    int rlen, first_time = 1;
    FILE *fname = NULL;

    /*
     * Read from the file system the file that contain the CA
     * certificate.  If the file does not exist, wait until it
     * is downloaded to the file system via the FTP server.
     */
    do
    {
        fname = fopen(fpath, "r");
        if (fname)
        {
            rlen = fread(buf, 1, buflen, fname);
            if ((!rlen) || (rlen == EOF))
            {
                printf("C library fread() returned 0, error %d\n", getErrno());
                rlen = 0;
            }
            break;
        }
        else
        {
            if (first_time)
            {
                printf("Please load %s through FTP...\n", fpath);
                first_time = 0;
            }
            tx_thread_sleep(NABspTicksPerSecond);
        }
    } while (fname == NULL);

    if (fclose(fname))
    {
        printf("C library fclose() failed\n");
    }

    return rlen;
}

/*
 *
 *  Function: static int ssh_shell_process(unsigned long)
 *
 *  Description:
 * 
 *      This function is a thread that runs a command 
 *      line instance over the file descriptor provided by
 *      the SSH shell callback.  This function could be 
 *      modified for other purposes, such as connecting 
 *      directly to a serial port, or an application.
 *
 *  Parameters:
 *
 *      unsigned long param : a pointer to an 
 *                            NaSsh2ProcessShellParams_t* structure
 *
 *  Return Values:
 *
 *      none 
 * 
 */ 
static void ssh_shell_process(unsigned long parm)
{
    NaSsh2ProcessShellParams_t* params = (NaSsh2ProcessShellParams_t*)parm;

    if(params == NULL)
    {
        ssh_app_send_cleanup();
        return;
    }

    /* Process CLI I/O */
    naCliCommandLoopDevice(params->shell_fd, NA_CLI_DEVICE_TYPE_SOCKET);
    closesocket(params->shell_fd);
    free(params);
    ssh_app_send_cleanup();
}
   
/*
 *
 *  Function: static int ssh_shell_callback(NaSsh2ProcessShellParams_t*)
 *
 *  Description:
 * 
 *      This function is called by the SSHv2 library when an SSHv2 client 
 *      requests a shell.  This is called from the context of the SSHv2 server,
 *      so it should not block indefinitely. In this example, a process is 
 *      spawned to run a command line on the shell file descriptor. 
 *      the shell_params parameter is owned by the v2 library, and the pointer 
 *      should not be used after returning from this function.
 *
 *  Parameters:
 *
 *      NaSsh2ProcessShellParams_t* : a pointer to an 
 *                                    NaSsh2ProcessShellParams_t* structure
 *                                    owned by the calling process.
 *
 *  Return Values:
 *
 *      int 0: if a shell was spawned, and a 
 *             channel was established.
 *          1: if a shell was not spawned, and
 *             no channel exists
 * 
 */  
int ssh_shell_callback(NaSsh2ProcessShellParams_t* shell_params)
{
    TX_THREAD* ssh_shell_process_thread;
    char* ssh_shell_process_stack;
    NaSsh2ProcessShellParams_t* shell_params_local;
    
    if(shell_params == NULL)
    {
        printf("Unable to spawn shell: invalid parameter\r\n");
        return -1;
    }

    shell_params_local = (NaSsh2ProcessShellParams_t*)malloc(sizeof(NaSsh2ProcessShellParams_t));

    if(shell_params_local == NULL)
    {
        printf("Unable to spawn shell: allocation failure\r\n");
        return -1;
    }

    memcpy(shell_params_local, shell_params, sizeof(NaSsh2ProcessShellParams_t));

    ssh_shell_process_thread = (TX_THREAD*)malloc(sizeof(TX_THREAD));

    if(ssh_shell_process_thread == NULL)
    {
        printf("Unable to spawn shell: unable to allocate thread");
        free(shell_params_local);
        return -1;
    }
 
    ssh_shell_process_stack = (char*)malloc(APP_ROOT_STACK_SIZE);

    if(ssh_shell_process_stack == NULL)
    {
        free(shell_params_local);
        free(ssh_shell_process_thread);
        return -1;
    } 

    /* Spawn a thread to process CLI commands */
    if(tx_thread_create(ssh_shell_process_thread,
                        "SSH CLI Shell Process",
                        ssh_shell_process,
                        (unsigned long)shell_params_local,
                        ssh_shell_process_stack,
                        APP_ROOT_STACK_SIZE,
                        APP_ROOT_PRIORITY,
                        APP_ROOT_PRIORITY,
                        1,
                        TX_AUTO_START) != TX_SUCCESS)
    {
        printf("Unable to spawn shell: unable to start processing thread");
        free(shell_params_local);
        free(ssh_shell_process_thread);
        free(ssh_shell_process_stack);
        return -1;
    } 
    else
    {
        return 0;
    }
}

/*
 *
 *  Function: static int ssh_connection_process(unsigned long)
 *
 *  Description:
 *      This function represents a thread spanwed from the applicationStart
 *      thread context that is used to service incoming SSHv2 connections.
 *      It simply passes in the NaSsh2ServerParams_t* params into the 
 *      NaSsh2ServerHandleIncoming() function and frees it when the 
 *      function completes the processing of the connection. The thread
 *      requests cleanup when it has completed its task.
 *
 *  Parameters:
 *
 *      unsigned long : a pointer to an 
 *                      NaSsh2ServerParams_t* structure
 *                      owned by this thread. This thread is
 *                      responsible for freeing this pointer.
 *
 *  Return Values:
 *
 *
 *      none
 * 
 */  
static void ssh_connection_process(unsigned long parm)
{

    NaSsh2ServerParams_t* server_params = (NaSsh2ServerParams_t*)parm;
    
    if(server_params == NULL)
    {
        ssh_app_send_cleanup();
        return;
    }

    /* Handle the incoming connection */
    NaSsh2ServerHandleIncoming(server_params);
    free(server_params);
    ssh_app_send_cleanup();
}


/* PEM formatted keys for evaluation purposes */
const char ssh_private_key_rsa[] =  
"-----BEGIN RSA PRIVATE KEY-----\n"
"Proc-Type: 4,ENCRYPTED\n"
"DEK-Info: DES-EDE3-CBC,16CA04B6DFAC6CF2\n"
"\n"
"G1gz6AABtryPR/9AbgKd+HpjKKOiL4ACiOGpLH7Rc1A76t+D67/jkJiS6W1SDcqH\n"
"Ielq40qaI7GPWGpXCG+XzmIWQlzXRK3Y15EXPeZVzz8aP+MNSgljjIONs+BhzDXw\n"
"ZtFND3xeE/MWZJ2hS0lakQPtcOCUWCof4hsyj5weevz8mVg8nVgXtrFzoMxzBfKF\n"
"N3eWKAncHUXYgUNb3y4b/EDKLApE3GTViMRmGi3gVD5JDdqm4Sgm+ixPTvbJpTWN\n"
"x5hsqv5Z+CZwm+miSELzFAA5klBWpk00qvq0o/RpQzmfXpsTImCkCor3ckgj6oGk\n"
"GsO6iOe1sf2sHgc3ahXNTVVxTxoWjfRg7STSjdt117qDVGwM5Ek1v5gvdQB7pv9L\n"
"wf+XiMzWbL1mkfhEYwRYvHqhXBDFomAaLSmN8VMqOPpU9L78zlRgbWKgZIIMcR0L\n"
"xMgz4h+s4rubwX48wkqHtg3cC5Jnot79CGEMEMc/H7YqIWgLtOnLLQ4lsMgzwOz8\n"
"uT6HcuAzETEmpnoQXB3/se2fRYEoV5ZSCu93YttfmxK7Sp+wpeHGf1uqToVw3Ac3\n"
"ofy54L4Vx+ZbwKG7mb2an2SW/hYIDys0kt8/0ZhmxFKq294YhIyIYfd/QIjknfRa\n"
"HYaDI5p3gQjxEBv3Z8l5cMUZDTCE/Npv+/yt9Fxcl4HSX2nHc29hK7MBV8GI32Pc\n"
"NC+C/b4cgYqQwXdWSEt5olwPX4A9X6RzHLMloDsN9toPDhyUswBJjpn9B7RI13li\n"
"pPzW11manR3/Yji/Ups/eauElWObK/C4ErRxRh1In+8=\n"
"-----END RSA PRIVATE KEY-----\n";

const char ssh_private_key_dsa[] =  
"-----BEGIN DSA PRIVATE KEY-----\n"
"Proc-Type: 4,ENCRYPTED\n"
"DEK-Info: DES-EDE3-CBC,D8EE2B1849184462\n"
"\n"
"eSO6PY8zBsnP1HLTHxt5WkVwoZ+cDvjwixpE+pjNeSMd9M7MdPWQdaNiSNJYJZq1\n"
"G7LNHBCQvBzoAtxlU9h7DKx8sZpIu8WAt1/hFYpGa5lsSvmM56e1efriyRy7DZiC\n"
"a2yCT0z5jzE4OyfRtkb7SNBKn21kxXT7ArXEySnksQ3nbJ493qci7/GF42cyj9C4\n"
"z3CvE5c4/b+8DLpVkvzq654KlggkykwZ4wA5k0kYZjVXzjKl/7gL/XlM4CYFuXh2\n"
"XMvU44iAopwyO+mlwPGu6H6U5ozcnalx6GCGJTRtVKmrI9jHQhq0KDYXk+vqeNAS\n"
"jM6RbrcTaFwQ17HvPQFKgMVd7t0W2cvvbo75uLi9MIzuQaaUiqyMtAqZMUDZPKF0\n"
"Tmw2hXSYw7Uqkx/my6zwR05Et+Fe9gJe41lRTKSYVHsoYqEPddwjVdvjp81PDt3X\n"
"bZCAyKN/r2VLHYfu2RKzgL20NAGSC12N8n50iO0QNDTtQ4ralzYggWEMOEGlzzSi\n"
"8PG9gSy8MeshOtLvUMatUvCdek4E29Fy350f13+bWG0Wn7RcnlGtTmp9jjjYA3k/\n"
"rLO564TtHHDmyR0cyGaSwg==\n"
"-----END DSA PRIVATE KEY-----\n";

/*
 *
 *  Function: void applicationStart (void)
 *
 *  Description:
 *
 *      This routine is responsible for starting the user application.  It should 
 *      create any threads or other resources the application needs.
 *
 *      ThreadX, the NET+OS device drivers, and the TCP/IP stack will be running
 *      when this function is called.
 *
 *  Parameters:
 *
 *      none
 *
 *  Return Values:
 *
 *      none
 * 
 */
void applicationStart (void)
{
    int i = 0;
    int sockfd = -1;
    struct sockaddr_storage serverAddr, peerAddr;
    int serverAddr_len, peerAddr_len;
    int server_socket;
    static char svr_key_buffer[MAX_SSH_KEY_LEN];
    int svr_key_len;
    int errorCode = 0;
    NaSsh2ServerParams_t* server_params;
    char* cleanup_stack;
    TX_THREAD* cleanup_thread;
    char* ssh_connection_process_stack;
    TX_THREAD* ssh_connection_process_thread;
    int nodelay = 1; 

    if (SHOULD_USE_STATIC_KEYS)
    {
        printf ("Using static example keys\n");
    }
    else /* Use FTP / filesystem */
    {
       /*
        * Link the FTP server to a file system volume and start the FTP server
        */
        /* Add the FTP server super user username and password */
        errorCode = NAsetSysAccess (NASYSACC_ADD, FTP_SUPERUSER_USERNAME, FTP_SUPERUSER_PASSWORD, NASYSACC_FS_GROUP2_READ | NASYSACC_FS_GROUP2_WRITE | NASYSACC_LEVEL_RW, NULL);
    
        if (errorCode != 0)
        {
            printf("NASetSysAccess Error: unsuccessfully set \"%s\" username \"%s\" password %d\n", 
                   FTP_SUPERUSER_USERNAME , FTP_SUPERUSER_PASSWORD, errorCode);
            goto exit;
        }

        /* Link FTP server to file system volume and start FTP server */
        errorCode = ftp_server_app_init();
        if (errorCode != 0)
        {
            printf("ftp_server_app_init Error: %d\n", errorCode);
            goto exit;
        }

        /* read the server key from data base */
        svr_key_len = readKeysFromDB(SVR_KEY_FILEPATH, svr_key_buffer, MAX_SSH_KEY_LEN);
        if (!svr_key_len)
        {
            printf("Unable to find hostkey %s in the filesystem\r\n", SVR_KEY_FILEPATH);
            goto exit;
        }
    }

    // set up the listener socket. If IPV6 is enabled then set family to ipv6. Ipv6 can listen for
	// both ipv4 adn ipv6

    #ifdef IPV6_ENABLED
	server_socket = socket (AF_INET6, SOCK_STREAM, 0);
	#else
    server_socket = socket (AF_INET, SOCK_STREAM, 0);
	#endif

    
    /*  Add username: digi password: sysadm to the System Access Database   */
    NAsetSysAccess (NASYSACC_ADD, SSH_USERNAME, SSH_PASSWORD, NASYSACC_LEVEL_SSH_RW, 0);

    setsockopt(server_socket, IPPROTO_TCP, TCP_NODELAY, (char *)&nodelay, sizeof(int));

    /* Listen for incoming SSHv2 connections on port 22 */
    memset(&serverAddr, 0, sizeof (serverAddr));

    #ifdef IPV6_ENABLED
    serverAddr.ss_family = AF_INET6;
	#else
    serverAddr.ss_family = AF_INET;
	#endif
    serverAddr.ss_port = htons(SSH_LISTEN_PORT);
    serverAddr_len = sizeof(serverAddr);

    if(bind(server_socket, (struct sockaddr *)&serverAddr, serverAddr_len) < 0)
    {
        printf("bind() encountered error %d.\n", getErrno());
        goto exit;
    }

    memset(&cleanup_queue, 0, sizeof(cleanup_queue));

    if(tx_queue_create(&cleanup_queue, "SSH Application Cleanup Queue", TX_4_ULONG, cleanup_queue_messages, CLEANUP_QUEUE_LENGTH) != TX_SUCCESS)
    {
        printf("Unable to create cleanup queue\r\n");
        goto exit;
    }

    /* Start the cleanup thread for removing spawned thread resources */
    cleanup_stack = (char*)malloc(CLEANUP_STACK_SIZE);

    if(cleanup_stack == NULL)
    {
        printf("Unable to allocate cleanup stack\r\n");
        goto exit;   
    }

    cleanup_thread = (TX_THREAD*)malloc(sizeof(TX_THREAD));

    if(cleanup_thread == NULL)
    {
        printf("Unable to allocate cleanup thread\r\n");
        goto exit; 
    }

    memset(cleanup_thread, 0, sizeof(TX_THREAD));

    if(tx_thread_create(cleanup_thread,
                        "SSH Application Resource Deallocator",
                        ssh_app_thread_cleanup,
                        (unsigned long)0,
                        cleanup_stack,
                        CLEANUP_STACK_SIZE,
                        APP_ROOT_PRIORITY,
                        APP_ROOT_PRIORITY,
                        1,
                        TX_AUTO_START) != TX_SUCCESS)
    {
        printf("Unable to create cleanup thread\r\n");
        goto exit;
    }

    if(listen (server_socket, 5) == 0)
    {
        peerAddr_len = sizeof(peerAddr);
    
        printf ("SSHv2 server is ready to accept incoming connections.\n");

        /* Loops forever, processing incoming SSHv2 client connections */
        for(;;)
        {
            printf("Waiting for new SSHv2 client to connect...\n");
            if((sockfd = accept(server_socket, (struct sockaddr *)&peerAddr, &peerAddr_len)) < 0)
                continue;

            setsockopt(sockfd, IPPROTO_TCP, TCP_NODELAY, (char *)&nodelay, sizeof(int));

            server_params = (NaSsh2ServerParams_t*)malloc(sizeof(NaSsh2ServerParams_t));

            if(server_params == NULL)
            {
                printf("SSHv2 incoming connection failed: allocation failure\n");
                continue;
            }

            /* Fill in SSH server parameters */
            server_params->fd = sockfd;
 
            if (SHOULD_USE_STATIC_KEYS)
            {
                server_params->host_key_rsa = (char*)ssh_private_key_rsa; 
                server_params->host_key_rsa_len = strlen(ssh_private_key_rsa);
                server_params->host_key_rsa_password = SVR_KEY_PASSWORD;
                server_params->host_key_dsa = (char*)ssh_private_key_dsa; 
                server_params->host_key_dsa_len = strlen(ssh_private_key_dsa);
                server_params->host_key_dsa_password = SVR_KEY_PASSWORD;
            }
            else
            {
                /* Single key, either DSA, or RSA, so populate both, and
                   let the API layer determine what type of key it is */
                server_params->host_key_rsa = svr_key_buffer; 
                server_params->host_key_rsa_len = svr_key_len;
                server_params->host_key_rsa_password = SVR_KEY_PASSWORD;
                server_params->host_key_dsa = svr_key_buffer; 
                server_params->host_key_dsa_len = svr_key_len;
                server_params->host_key_dsa_password = SVR_KEY_PASSWORD;
            }

            /* Initialize server PRNG seed. */
            for(i=0;i<SSH_RANDOM_SEED_LEN;i++)
                server_params->random_seed[i] = tx_time_get() + 1973*i;

            /* Function to call when the SSHv2 client requests a shell */
            server_params->shell_callback = ssh_shell_callback;  
    
            /* Start the cleanup thread for removing spawned thread resources */
            ssh_connection_process_stack = (char*)malloc(APP_ROOT_STACK_SIZE);

            if(ssh_connection_process_stack == NULL)
            {
                printf("Unable to allocate cleanup stack\r\n"); 
                free(server_params);
                continue;
            }   
            
            ssh_connection_process_thread = (TX_THREAD*)malloc(sizeof(TX_THREAD));

            if(ssh_connection_process_thread == NULL)
            {
                printf("Unable to allocate cleanup thread\r\n");
                free(server_params);
                free(ssh_connection_process_stack);
                continue;
            }

            memset(ssh_connection_process_thread, 0, sizeof(TX_THREAD));

            /* Spawn a thread to process the incoming connection */
            if(tx_thread_create(ssh_connection_process_thread,
                                "SSH Connection Process",
                                ssh_connection_process,
                                (unsigned long)server_params,
                                ssh_connection_process_stack,
                                APP_ROOT_STACK_SIZE,
                                APP_ROOT_PRIORITY,
                                APP_ROOT_PRIORITY,
                                1,
                                TX_AUTO_START) != TX_SUCCESS)
            {
                free(server_params);
                free(ssh_connection_process_stack);
                free(ssh_connection_process_thread);
                printf("SSHv2 incoming connection failed: unable to spawn thread\n");
                closesocket(sockfd);
            } 
            else
            {
                printf("SSHv2 spawned process to handle incoming connection...\n");
            }
        }
    }
    else
    {
        printf ("listen() encountered error %d.\n", getErrno());
    }

exit:
    if (server_socket >= 0) closesocket(server_socket);    
    
    tx_thread_suspend(tx_thread_identify());
}

/*
 *
 *  Function: void applicationTcpDown (void)
 *
 *  Description:
 * 
 *      This routine will be called by the NET+OS root thread once every 
 *      clock tick while it is waiting for the TCP/IP stack to come up.  
 *      This function can increment a counter everytime it's called to 
 *      keep track of how long we've been waiting for the stack to start.
 *      If we've been waiting too long, then this function can do something
 *      to handle the error.  
 *
 *      This function will not be called once the stack has started.
 *
 *  Parameters:
 *
 *      none
 *
 *  Return Values:
 *
 *      none
 *
 */

void applicationTcpDown (void)

{
    static int ticksPassed = 0;

    ticksPassed++;
/*
 * Code to handle error condition if the stack doesn't come up goes here.
 */
}

/* Required for linking */
int HSInitSecurityTable() { return 0; }
int SearchURL (unsigned long nau, char * nac) { return 0; }
